from collections.abc import Iterable
from typing import Literal, Optional, TypeVar, Union
from warnings import warn

import casadi as cs
import numpy as np
import numpy.typing as npt

from ..core.scaling import Scaler
from ..core.solutions import Solution, subsevalf
from .wrapper import Nlp, NonRetroactiveWrapper

SymType = TypeVar("SymType", cs.SX, cs.MX)


class NlpScaling(NonRetroactiveWrapper[SymType]):
    """
    Wraps an NLP problem to facilitate the scaling of parameters and variables as well
    as the automatic scaling of expression (e.g., objective and constraints).
    """

    def __init__(self, nlp: Nlp[SymType], scaler: Scaler, warns: bool = True) -> None:
        """Initializes a scaling wrapper around an NLP instance.

        Parameters
        ----------
        nlp : Nlp[T]
            The NLP problem to be wrapped.
        scaler : Scaler
            A class for scaling the NLP's quantities. See `csnlp.core.scaling` for a
            collection of these.
        warns : bool, optional
            If `True`, warns each time a variable or parameter is created which has not
            been registered to the scaler and thus cannot be scaled; otherwise, it will
            not raise warnings.
        """
        super().__init__(nlp)
        self.scaler = scaler
        self.warns = warns
        self._svars: dict[str, SymType] = {}
        self._spars: dict[str, SymType] = {}
        self._uvars: dict[str, SymType] = {}
        self._upars: dict[str, SymType] = {}

    @property
    def scaled_variables(self) -> dict[str, SymType]:
        """Gets the scaled variables of the NLP scheme."""
        return self._svars

    @property
    def scaled_parameters(self) -> dict[str, SymType]:
        """Gets the scaled parameters of the NLP scheme."""
        return self._spars

    @property
    def unscaled_variables(self) -> dict[str, SymType]:
        """Gets the unscaled variables of the NLP scheme."""
        return self._uvars

    @property
    def unscaled_parameters(self) -> dict[str, SymType]:
        """Gets the unscaled parameters of the NLP scheme."""
        return self._upars

    def scale(self, expr: SymType) -> SymType:
        """Scales an expression with the MPC's scaled variables and parameters.

        Parameters
        ----------
        expr : casadi.SX or MX
            The expression to be scaled.

        Returns
        -------
        casadi.SX or MX
            The scaled expression.
        """
        expr = subsevalf(expr, self.nlp.variables, self._svars, eval=False)
        expr = subsevalf(expr, self.nlp.parameters, self._spars, eval=False)
        return expr

    def unscale(self, expr: SymType) -> SymType:
        """Unscales an expression with the MPC's unscaled variables and parameters.

        Parameters
        ----------
        expr : casadi.SX or MX
            The expression to be unscaled.

        Returns
        -------
        casadi.SX or MX
            The unscaled expression.
        """
        expr = subsevalf(expr, self.nlp.variables, self._uvars, eval=False)
        expr = subsevalf(expr, self.nlp.parameters, self._upars, eval=False)
        return expr

    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        """See `Nlp.variable` method."""
        can_scale = name in self.scaler
        if can_scale:
            lb, ub = np.broadcast_to(lb, shape), np.broadcast_to(ub, shape)
            lb = self.scaler.scale(name, lb)
            ub = self.scaler.scale(name, ub)
        var, lam_lb, lam_ub = self.nlp.variable(name, shape, lb, ub)
        if can_scale:
            svar = self.scaler.scale(name, var)
            uvar = self.scaler.unscale(name, var)
        else:
            if self.warns:
                warn(f"Scaling for variable {name} not found.", RuntimeWarning)
            svar = uvar = var
        self._svars[name] = svar
        self._uvars[name] = uvar
        return var, lam_lb, lam_ub

    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        """See `Nlp.parameter` method."""
        par = self.nlp.parameter(name, shape)
        if name in self.scaler:
            spar = self.scaler.scale(name, par)
            upar = self.scaler.unscale(name, par)
        else:
            if self.warns:
                warn(f"Scaling for parameter {name} not found.", RuntimeWarning)
            spar = upar = par
        self._spars[name] = spar
        self._upars[name] = upar
        return par

    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        """See `Nlp.constraint` method."""
        return self.nlp.constraint(name, self.unscale(lhs - rhs), op, 0, soft, simplify)

    def minimize(self, objective: SymType) -> None:
        """See `Nlp.minimize` method."""
        return self.nlp.minimize(self.unscale(objective))

    def solve(
        self,
        pars: Optional[dict[str, npt.ArrayLike]] = None,
        vals0: Optional[dict[str, npt.ArrayLike]] = None,
    ) -> Solution[SymType]:
        """See `Nlp.solve` method."""
        if pars is not None:
            pars = self._scale_dict(pars)
        if vals0 is not None:
            vals0 = self._scale_dict(vals0)
        return self.nlp.solve(pars, vals0)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        return_stacked_sol: bool = False,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        """See `MultistartNlp.solve` method."""
        assert self.nlp.is_multi and hasattr(
            self.nlp, "solve_multi"
        ), "`solve_multi` called on an nlp instance that is not `MultistartNlp`."
        if pars is not None:
            pars = (
                self._scale_dict(pars)
                if isinstance(pars, dict)
                else (self._scale_dict(pars_i) for pars_i in pars)
            )
        if vals0 is not None:
            vals0 = (
                self._scale_dict(vals0)
                if isinstance(vals0, dict)
                else (self._scale_dict(vals0_i) for vals0_i in vals0)
            )
        return self.nlp.solve_multi(
            pars,
            vals0,
            return_all_sols=return_all_sols,
            return_stacked_sol=return_stacked_sol,
        )

    def _scale_dict(self, d: dict[str, npt.ArrayLike]) -> dict[str, npt.ArrayLike]:
        """Internal utility for scaling structures/dicts."""
        scaler = self.scaler
        return {
            name: scaler.scale(name, d[name]) if scaler.can_scale(name) else d[name]
            for name in d
        }



from collections.abc import Iterable, Iterator
from functools import lru_cache, partial
from itertools import repeat
from typing import Any, ClassVar, Generic, Literal, Optional, TypeVar, Union

import casadi as cs
import numpy as np
import numpy.typing as npt
from joblib import Parallel, delayed
from joblib.memory import MemorizedFunc

from ..core.cache import invalidate_cache
from ..core.solutions import Solution, subsevalf
from ..nlps.nlp import Nlp
from ..nlps.objective import _solve_and_get_stats

SymType = TypeVar("SymType", cs.SX, cs.MX)


def _n(sym_name: str, scenario: int) -> str:
    """Internal utility for the naming convention of i-scenario's symbols."""
    return f"{sym_name}__{scenario}"


def _chained_subevalf(
    expr: Union[SymType, np.ndarray],
    old_vars: dict[str, SymType],
    new_vars: dict[str, SymType],
    old_pars: dict[str, SymType],
    new_pars: dict[str, SymType],
    old_dual_vars: Optional[dict[str, SymType]] = None,
    new_dual_vars: Optional[dict[str, SymType]] = None,
    eval: bool = True,
) -> Union[SymType, cs.DM, np.ndarray]:
    """Internal utility to perform `subevalf` on vars, pars and duals in chain."""
    expr = subsevalf(expr, old_vars, new_vars, eval=False)
    if old_dual_vars is None:
        return subsevalf(expr, old_pars, new_pars, eval=eval)
    expr = subsevalf(expr, old_pars, new_pars, eval=False)
    return subsevalf(expr, old_dual_vars, new_dual_vars, eval=eval)


def _find_best_sol(sols: Iterator[dict[str, Any]]) -> dict[str, Any]:
    """Picks the best solution out of multiple solutions, with the following logic: the
    current solution should be considered better if
     * it is feasible and the best is not, or
     * both are feasible or infeasible, and the current is successful and the best is
       not, or
     * both are successful or not, and the current has a lower f than the best.
    """
    best_sol = next(sols)
    best_f = float(best_sol["f"])
    is_best_successful = best_sol["stats"]["success"]
    is_best_feasible = "infeasib" not in best_sol["stats"]["return_status"].lower()
    for sol in sols:
        f = float(sol["f"])
        is_successful = sol["stats"]["success"]
        is_feasible = "infeasib" not in sol["stats"]["return_status"].lower()
        if (
            (is_feasible and not is_best_feasible)
            or (
                is_feasible == is_best_feasible
                and (is_successful and not is_best_successful)
            )
            or (is_successful == is_best_successful and f < best_f)
        ):
            best_sol = sol
            best_f = f
            is_best_successful = is_successful
            is_best_feasible = is_feasible
    return best_sol


class MultistartNlp(Nlp[SymType], Generic[SymType]):
    """Base class for NLP with multistarting."""

    is_multi: ClassVar[bool] = True

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        """Initializes the multistart NLP instance.

        Parameters
        ----------
        args, kwargs
            See inherited `csnlp.Nlp`.
        starts : int
            A positive integer for the number of multiple starting guesses to optimize.

        Raises
        ------
        ValueError
            Raises if the scenario number is invalid.
        """
        if starts <= 0:
            raise ValueError("Number of scenarios must be positive and > 0.")
        super().__init__(*args, **kwargs)
        self._starts = starts

    def __call__(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        **kwargs: Any,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        # call solve_multi only if either pars or vals0 is an iterable; otherwise, run
        # the single, base NLP
        if (pars is None or isinstance(pars, dict)) and (
            vals0 is None or isinstance(vals0, dict)
        ):
            return self.solve(pars, vals0)
        return self.solve_multi(pars, vals0, **kwargs)

    @property
    def starts(self) -> int:
        """Gets the number of starts."""
        return self._starts

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        """Solves the NLP with multiple initial conditions.

        Parameters
        ----------
        pars : dict[str, array_like] or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            parameter in the NLP scheme, the corresponding numerical value. In case a
            single dict is passed, the same is used across all scenarions. Can be `None`
            if no parameters are present.
        vals0 : dict[str, array_like] or iterable of, optional
            An iterable that, for each multistart, contains a dictionary with, for each
            variable in the NLP scheme, the corresponding initial guess. In case a
            single dict is passed, the same is used across all scenarions. By default
            `None`, in which case  initial guesses are not passed to the solver.
        return_all_sols : bool, optional
            If `True`, returns the solution of each multistart of the NLP; otherwise,
            only the best solution is returned. By default, `False`.

        Returns
        -------
        Solution or list of Solutions
            Depending on the flags `return_all_sols`, returns
             - the best solution out of all multiple starts
             - all the solutions (one per start)
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement `solve_multi`"
        )


class StackedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that models and solves an NLP from multiple starting initial guesses by
    automatically stacking the problem multiple independent times in the same,
    larger-scale NLP. This allows to solve the original problem multiple times via a
    single call to the solver."""

    def __init__(self, *args: Any, starts: int, **kwargs: Any) -> None:
        # this class essentially is a facade that hides an internal nlp in which the
        # problem (variables, parameters, etc.) are duplicated by the requested number
        # of multiple starts. For this reason, all methods are overridden to create
        # multiples of these in the hidden nlp.
        super().__init__(*args, starts=starts, **kwargs)
        self._stacked_nlp = Nlp(*args, **kwargs)  # actual nlp

    @lru_cache(maxsize=32)
    def _vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the variables of the i-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._vars[_n(k, i)] for k in self._vars}

    @lru_cache(maxsize=32)
    def _pars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the parameters of the i-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._pars[_n(k, i)] for k in self._pars}

    @lru_cache(maxsize=32)
    def _dual_vars_i(self, i: int) -> dict[str, SymType]:
        """Internal utility to retrieve the dual variables of the i-th scenario."""
        nlp = self._stacked_nlp.unwrapped
        return {k: nlp._dual_vars[_n(k, i)] for k in self._dual_vars}

    @invalidate_cache(_pars_i)
    def parameter(self, name: str, shape: tuple[int, int] = (1, 1)) -> SymType:
        out = super().parameter(name, shape)
        for i in range(self._starts):
            self._stacked_nlp.parameter(_n(name, i), shape)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def variable(
        self,
        name: str,
        shape: tuple[int, int] = (1, 1),
        lb: Union[npt.ArrayLike, cs.DM] = -np.inf,
        ub: Union[npt.ArrayLike, cs.DM] = +np.inf,
    ) -> tuple[SymType, SymType, SymType]:
        out = super().variable(name, shape, lb, ub)
        for i in range(self._starts):
            self._stacked_nlp.variable(_n(name, i), shape, lb, ub)
        return out

    @invalidate_cache(_vars_i, _dual_vars_i)
    def constraint(
        self,
        name: str,
        lhs: Union[SymType, np.ndarray, cs.DM],
        op: Literal["==", ">=", "<="],
        rhs: Union[SymType, np.ndarray, cs.DM],
        soft: bool = False,
        simplify: bool = True,
    ) -> tuple[SymType, ...]:
        expr = lhs - rhs
        if simplify:
            expr = cs.cse(cs.simplify(expr))
        out = super().constraint(name, expr, op, 0, soft, False)

        # NOTE: the line above already created the slack variables in each scenario, so
        # below we have to pass both soft=False and the expression with slack included.
        vars = self.variables
        pars = self.parameters
        expr_ = out[0]  # slack-relaxed expression in the form h(x)<=0 or ==0
        op_: Literal["==", "<="] = "==" if op == "==" else "<="
        for i in range(self._starts):
            expr_i = _chained_subevalf(
                expr_, vars, self._vars_i(i), pars, self._pars_i(i), eval=False
            )
            self._stacked_nlp.constraint(_n(name, i), expr_i, op_, 0, False, False)
        return out

    def minimize(self, objective: SymType) -> None:
        out = super().minimize(objective)

        vars = self.variables
        pars = self.parameters
        self._fs: list[SymType] = [
            _chained_subevalf(
                objective, vars, self._vars_i(i), pars, self._pars_i(i), eval=False
            )
            for i in range(self._starts)
        ]
        self._stacked_nlp.minimize(sum(self._fs))
        return out

    def init_solver(
        self,
        opts: Optional[dict[str, Any]] = None,
        solver: str = "ipopt",
        type: Literal["auto", "nlp", "conic"] = "auto",
    ) -> None:
        out = super().init_solver(opts, solver, type)
        self._stacked_nlp.init_solver(opts, solver, type)
        return out

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        return_stacked_sol: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (
            return_stacked_sol and return_all_sols
        ), "`return_all_sols` and `return_stacked_sol` can't be both true."
        if pars is not None:
            pars_iter = repeat(pars, self.starts) if isinstance(pars, dict) else pars
            pars = {
                _n(n, i): pars_i[n]
                for i, pars_i in enumerate(pars_iter)
                for n in pars_i.keys()
            }
        if vals0 is not None:
            v0_iter = repeat(vals0, self.starts) if isinstance(vals0, dict) else vals0
            vals0 = {
                _n(n, i): vals0_i[n]
                for i, vals0_i in enumerate(v0_iter)
                for n in vals0_i.keys()
            }
        multi_sol = self._stacked_nlp.solve(pars, vals0)
        if return_stacked_sol:
            return multi_sol

        vars_ = self.variables
        pars_ = self.parameters
        duals_ = self.dual_variables
        old = cs.vertcat(
            self._x, self._lam_g, self._lam_h, self._lam_lbx, self._lam_ubx, self._p
        )
        fs = [float(multi_sol._get_value(f)) for f in self._fs]

        def get_ith_sol(idx: int) -> Solution[SymType]:
            vals = {n: multi_sol.vals[_n(n, idx)] for n in vars_.keys()}
            dual_vals = {n: multi_sol.dual_vals[_n(n, idx)] for n in duals_.keys()}
            new = multi_sol._get_value(
                _chained_subevalf(
                    old,
                    vars_,
                    self._vars_i(idx),
                    pars_,
                    self._pars_i(idx),
                    duals_,
                    self._dual_vars_i(idx),
                    False,
                )
            )
            get_value = partial(subsevalf, old=old, new=new)
            return Solution(
                fs[idx], vars_, vals, duals_, dual_vals, multi_sol.stats, get_value
            )

        if return_all_sols:
            return [get_ith_sol(i) for i in range(self._starts)]
        return get_ith_sol(np.argmin(fs).item())

    def remove_variable_bounds(
        self,
        name: str,
        direction: Literal["lb", "ub", "both"],
        idx: Union[tuple[int, int], list[tuple[int, int]]] = None,
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_variable_bounds(name, direction, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_variable_bounds(_n(name, i), direction, idx)

    def remove_constraints(
        self, name: str, idx: Union[tuple[int, int], list[tuple[int, int]]] = None
    ) -> None:
        idx = [idx] if isinstance(idx, tuple) else list(idx)
        super().remove_constraints(name, idx)
        for i in range(self._starts):
            self._stacked_nlp.remove_constraints(_n(name, i), idx)


class ParallelMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP via parallelization of the computations."""

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallel: Optional[Parallel] = None,
        n_jobs: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the multistart NLP instance.

        Parameters
        ----------
        args, kwargs
            See inherited `csnlp.Nlp`.
        starts : int
            A positive integer for the number of multiple starting guesses to optimize.
        parallel: joblib.Parallel, optional
            A parallel backend to use (must be already initialized); if `None`, a new
            one is instead created and initialized.
        n_jobs : int, optional
            Number of concurrently running jobs; see `n_job` in `joblib.Parallel`. Only
            used if the `parallel` backend is not provided.

        Raises
        ------
        ValueError
            Raises if the scenario number is invalid.
        """
        super().__init__(*args, starts=starts, **kwargs)
        if parallel is None:
            self._parallel = Parallel(n_jobs=n_jobs, return_as="generator_unordered")
            self.initialize_parallel()
        else:
            self._parallel = parallel
        self._n_jobs = self._parallel.n_jobs

    def initialize_parallel(self) -> None:
        """Initializes the parallel backend."""
        self._parallel.__enter__()

    def terminate_parallel(self) -> None:
        """Terminates the parallel backend."""
        self._parallel.__exit__(None, None, None)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        if self._solver is None:
            raise RuntimeError("Solver uninitialized.")
        shared_kwargs = {
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        kwargs = (
            self._process_pars_and_vals0(shared_kwargs.copy(), p, v0)
            for p, v0 in zip(pars_iter, vals0_iter)
        )
        sols: Iterator[dict[str, Any]] = self._parallel(
            delayed(_solve_and_get_stats)(self._solver, kw) for kw in kwargs
        )
        if return_all_sols:
            return list(map(self._process_solver_sol, sols))
        best_sol = _find_best_sol(sols)
        return self._process_solver_sol(best_sol)

    def __getstate__(self, fullstate: bool = False) -> dict[str, Any]:
        # joblib.Parallel cannot be pickled or deepcopied
        state = super().__getstate__(fullstate)
        state.pop("_parallel", None)
        return state

    def __setstate__(self, state: Optional[dict[str, Any]]) -> None:
        if state is not None:
            self.__dict__.update(state)
        # re-initialized joblib.Parallel
        if not hasattr(self, "_n_jobs"):
            self._n_jobs = None
        self._parallel = Parallel(n_jobs=self._n_jobs, return_as="generator_unordered")
        self.initialize_parallel()


class MappedMultistartNlp(MultistartNlp[SymType], Generic[SymType]):
    """A class that solves an NLP multiple times in parallel via `casadi.Function.map`
    parallelization.

    See https://github.com/casadi/casadi/wiki/FAQ:-How-to-use-map%28%29-and-mapaccum%28%29-to-speed-up-calculations%3F.
    """

    def __init__(
        self,
        *args: Any,
        starts: int,
        parallelization: Literal[
            "serial", "unroll", "inline", "thread", "openmp"
        ] = "serial",
        max_num_threads: Optional[int] = None,
        **kwargs: Any,
    ) -> None:
        """Initializes the multistart NLP instance.

        Parameters
        ----------
        args, kwargs
            See inherited `csnlp.Nlp`.
        starts : int
            A positive integer for the number of multiple starting guesses to optimize.
        parallelization : "serial", "unroll", "inline", "thread", "openmp"
            The type of parallelization to use (see `casadi.Function.map`). By default,
            `"serial"` is selected.
        max_num_threads : int, optional
            Maximum number of threads to use in parallelization; if `None`, the number
            of threads is equal to the number of starts.

        Raises
        ------
        ValueError
            Raises if the scenario number is invalid.
        """
        super().__init__(*args, starts=starts, **kwargs)
        self._mapped_solver: Optional[MemorizedFunc] = None
        self._parallelization = parallelization
        self._max_num_threads = max_num_threads

    def init_solver(self, *args: Any, **kwargs: Any) -> None:
        super().init_solver(*args, **kwargs)
        solver: cs.Function = self._solver.func
        mapped_solver = solver.map(
            self._starts, self._parallelization, self._max_num_threads or self._starts
        )
        self._mapped_solver = self._cache.cache(mapped_solver)

    def solve_multi(
        self,
        pars: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        vals0: Union[
            None, dict[str, npt.ArrayLike], Iterable[dict[str, npt.ArrayLike]]
        ] = None,
        return_all_sols: bool = False,
        _return_mapped_sol: bool = False,
        **_,
    ) -> Union[Solution[SymType], list[Solution[SymType]]]:
        assert not (
            _return_mapped_sol and return_all_sols
        ), "`return_all_sols` and `_return_mapped_sol` can't be both true."
        if self._mapped_solver is None:
            raise RuntimeError("Solver uninitialized.")
        pars_iter = (
            repeat(pars, self.starts)
            if pars is None or isinstance(pars, dict)
            else pars
        )
        vals0_iter = (
            repeat(vals0, self.starts)
            if vals0 is None or isinstance(vals0, dict)
            else vals0
        )
        x0s = []
        ps = []
        default = cs.DM.zeros(0, 1)
        for p, v0 in zip(pars_iter, vals0_iter):
            kwargs = self._process_pars_and_vals0({}, p, v0)
            x0s.append(kwargs.get("x0", default))
            ps.append(kwargs.get("p", default))
        single_kwargs = {
            "x0": cs.hcat(x0s),
            "p": cs.hcat(ps),
            "lbx": self._lbx.data,
            "ubx": self._ubx.data,
            "lbg": np.concatenate((np.zeros(self.ng), np.full(self.nh, -np.inf))),
            "ubg": 0,
        }
        single_sol: dict[str, cs.DM] = self._mapped_solver(**single_kwargs)

        # NOTE: the mapped solver does not return the stats, so we have to use the
        # original solver to get them - this likely returns the stats of the last
        # run of the mapped solver, but it's the best we can do up to this point
        stats = self._solver.func.stats()

        # if the user wants the mapped solution, return it, though it's not a Solution
        # and it's mostly for debugging
        if _return_mapped_sol:
            single_sol["p"] = single_kwargs["p"]
            single_sol["stats"] = stats
            return single_sol  # NOTE: this is NOT a Solution object

        # convert the mapped solution to a list of Solution objects
        if return_all_sols:
            sols = []
            for i, p in enumerate(ps):
                sol_i = {k: v[:, i] for k, v in single_sol.items()}
                sol_i["p"] = ps[i]
                sol_i["stats"] = stats.copy()
                sols.append(self._process_solver_sol(sol_i))
            return sols

        # just return the best, as we cannot use stats to exclude failed solutions
        i = np.argmin(single_sol["f"])
        sol = {k: v[:, i] for k, v in single_sol.items()}
        sol["p"] = ps[i]
        sol["stats"] = stats
        return self._process_solver_sol(sol)



import random
import unittest
from functools import cached_property, lru_cache, partial
from itertools import product
from typing import Union

import casadi as cs
import numpy as np
from casadi.tools import entry, struct_MX, struct_SX
from parameterized import parameterized

from csnlp.core.cache import invalidate_cache
from csnlp.core.data import array2cs, cs2array, find_index_in_vector
from csnlp.core.debug import NlpDebug, NlpDebugEntry
from csnlp.core.derivatives import hohessian, hojacobian
from csnlp.core.scaling import MinMaxScaler, Scaler
from csnlp.core.solutions import Solution, subsevalf

GROUPS = set(NlpDebug.types.keys())


class Dummy:
    def __init__(self) -> None:
        self.counter1 = 0
        self.counter2 = 0

    @cached_property
    def prop1(self) -> int:
        self.counter1 += 1
        return self.counter1

    @lru_cache
    def method2(self) -> int:
        self.counter2 += 1
        return self.counter2

    @invalidate_cache(prop1, method2)
    def clear_cache(self) -> None:
        return


class Dummy2(Dummy):
    def __init__(self) -> None:
        super().__init__()
        self.counter3 = 0

    @cached_property
    def prop3(self) -> int:
        self.counter3 += 1
        return self.counter3

    @invalidate_cache(prop3)
    def clear_cache(self) -> None:
        return super().clear_cache()


class TestFuncs(unittest.TestCase):
    def test_invalidate_cache__raises__with_invalid_type(self):
        with self.assertRaises(TypeError):
            invalidate_cache(5)

    def test_invalidate_cache__clears_property_cache(self):
        # sourcery skip: extract-duplicate-method
        dummy = Dummy()
        dummy.prop1
        dummy.prop1
        dummy.method2()
        dummy.method2()
        self.assertEqual(dummy.counter1, 1)
        self.assertEqual(dummy.counter2, 1)
        dummy.clear_cache()
        dummy.prop1
        dummy.method2()
        self.assertEqual(dummy.counter1, 2)
        self.assertEqual(dummy.counter2, 2)
        dummy.clear_cache()
        dummy.prop1
        dummy.prop1
        dummy.method2()
        dummy.method2()
        self.assertEqual(dummy.counter1, 3)
        self.assertEqual(dummy.counter2, 3)

    def test_invalidate_cache__invalidates_only_current_object(self):
        # sourcery skip: extract-duplicate-method
        dummy1, dummy2 = Dummy(), Dummy()
        dummy1.prop1
        dummy2.prop1
        self.assertEqual(dummy1.counter1, 1)
        self.assertEqual(dummy2.counter1, 1)
        dummy1.clear_cache()
        dummy1.prop1
        dummy2.prop1
        self.assertEqual(dummy1.counter1, 2)
        self.assertEqual(dummy2.counter1, 1)

    def test_invalidate_cache__accepts_new_caches_to_clear(self):
        # sourcery skip: extract-duplicate-method
        dummy = Dummy2()
        dummy.prop1
        dummy.prop1
        dummy.method2()
        dummy.method2()
        dummy.prop3
        dummy.prop3
        self.assertEqual(dummy.counter1, 1)
        self.assertEqual(dummy.counter2, 1)
        self.assertEqual(dummy.counter3, 1)
        dummy.clear_cache()
        dummy.prop1
        dummy.method2()
        dummy.prop3
        self.assertEqual(dummy.counter1, 2)
        self.assertEqual(dummy.counter2, 2)
        self.assertEqual(dummy.counter3, 2)


class TestNlpDebug(unittest.TestCase):
    def test_register__adds_correct_info(self):
        debug = NlpDebug()
        name = "a name"
        shape = (2, 3)
        for group in GROUPS:
            debug.register(group, name, shape)
            info: tuple[range, NlpDebugEntry] = getattr(debug, f"_{group}_info")[0]
            self.assertEqual(info[0], range(shape[0] * shape[1]))
            self.assertEqual(info[1].name, name)
            self.assertEqual(info[1].shape, shape)
            self.assertEqual(info[1].type, NlpDebug.types[group])

    def test_register__raises__with_invalid_group(self):
        debug = NlpDebug()
        while True:
            group = chr(random.randint(ord("a"), ord("z")))
            if group not in GROUPS:
                break
        with self.assertRaises(AttributeError):
            debug.register(group, "var1", (1, 2))

    def test_xhg_describe__gets_corret_variables(self):
        debug = NlpDebug()
        for group in GROUPS:
            debug.register(group, "var1", (3, 3))
            debug.register(group, "var2", (1, 1))
            info1: NlpDebugEntry = getattr(debug, f"{group}_describe")(0)
            info2: NlpDebugEntry = getattr(debug, f"{group}_describe")(9)
            self.assertEqual(info1.name, "var1")
            self.assertEqual(info2.name, "var2")

    def test_xhg_describe__raises__with_outofbound_index(self):
        debug = NlpDebug()
        for group in GROUPS:
            debug.register(group, "var1", (1, 1))
            debug.register(group, "var2", (1, 1))
            with self.assertRaises(IndexError):
                getattr(debug, f"{group}_describe")(10_000)


class TestSolutions(unittest.TestCase):
    def test_subsevalf__raises__when_type_is_invalid(self):
        with self.assertRaises(Exception):
            subsevalf(None, unittest.TestCase, None)

    def test_subsevalf__raises__when_evaluating_a_symbolic_expr(self):
        x = cs.SX.sym("x")
        y = cs.SX.sym("y")
        with self.assertRaises(RuntimeError):
            subsevalf(x**2, x, y, eval=True)

    @parameterized.expand([(cs.SX, struct_SX), (cs.MX, struct_MX)])
    def test_subsevalf__computes_correct_value(
        self,
        XX: Union[type[cs.SX], type[cs.MX]],
        struct_X: Union[type[struct_SX], type[struct_MX]],
    ):
        shape = (3, 4)

        V = {
            "x": (XX.sym("x", *shape), np.random.rand(*shape) * 10),
            "y": (XX.sym("y", *shape), np.random.rand(*shape) * 5),
            "z": (XX.sym("z"), np.random.rand() + 1),
        }
        V_vec = cs.vertcat(*(cs.vec(v) for _, v in V.values()))
        V_struct = struct_X([entry(n, expr=s) for n, (s, _) in V.items()])
        expr, expected_val = ((V["x"][i] / V["y"][i]) ** V["z"][i] for i in range(2))

        actual_vals = [
            subsevalf(
                subsevalf(
                    subsevalf(expr, V["x"][0], V["x"][1], eval=False),
                    V["y"][0],
                    V["y"][1],
                    eval=False,
                ),
                V["z"][0],
                V["z"][1],
                eval=True,
            ),
            subsevalf(
                expr,
                (s for s, _ in V.values()),
                (v for _, v in V.values()),
                eval=True,
            ),
            subsevalf(
                expr,
                {n: s for n, (s, _) in V.items()},
                {n: v for n, (_, v) in V.items()},
                eval=True,
            ),
            subsevalf(expr, V_struct, V_struct(V_vec), eval=True),
        ]
        for actual_val in actual_vals:
            np.testing.assert_allclose(expected_val, actual_val)

    @parameterized.expand([(cs.SX, struct_SX), (cs.MX, struct_MX)])
    def test_solution__computes_correct_value(
        self,
        XX: Union[type[cs.SX], type[cs.MX]],
        struct_X: Union[type[struct_SX], type[struct_MX]],
    ):
        f = 0
        shape = (3, 4)
        V = {
            "x": (XX.sym("x", *shape), np.random.rand(*shape) * 10),
            "y": (XX.sym("y", *shape), np.random.rand(*shape) * 5),
            "z": (XX.sym("z"), np.random.rand() + 1),
        }
        D = {
            "dx": (XX.sym("dx", *shape), np.random.rand(*shape) * 10),
            "dy": (XX.sym("dy", *shape), np.random.rand(*shape) * 5),
            "dz": (XX.sym("dz"), np.random.rand() + 1),
        }
        V_vec = cs.vertcat(*(cs.vec(v) for _, v in V.values()))
        D_vec = cs.vertcat(*(cs.vec(d) for _, d in D.values()))
        V_struct = struct_X([entry(n, expr=s) for n, (s, _) in V.items()])
        D_struct = struct_X([entry(n, expr=s) for n, (s, _) in D.items()])

        expr, expected_val = (
            ((V["x"][i] + D["dx"][i]) / V["y"][i] / D["dy"][i])
            ** (V["z"][i] - D["dz"][i])
            for i in range(2)
        )

        all_vars = cs.vertcat(V_struct, D_struct)
        all_vals = cs.vertcat(V_struct(V_vec), D_struct(D_vec))
        get_value = partial(subsevalf, old=all_vars, new=all_vals)
        S = Solution(
            f=f,
            vars=V_struct,
            vals=V_struct(V_vec),
            dual_vars=D_struct,
            dual_vals=D_struct(V_vec),
            stats={},
            _get_value=get_value,
        )
        np.testing.assert_allclose(expected_val, S.value(expr))

    @parameterized.expand([(False,), (True,)])
    def test_solution__reports_success_and_barrier_properly(self, flag: bool):
        mu = np.abs(np.random.randn(10)).tolist()
        S = Solution(
            f=None,
            vars=None,
            vals=None,
            dual_vars=None,
            dual_vals=None,
            stats={"success": flag, "iterations": {"mu": mu}},
            _get_value=lambda x: x,
        )
        self.assertEqual(S.success, flag)
        self.assertEqual(S.barrier_parameter, mu[-1])


class TestData(unittest.TestCase):
    @parameterized.expand(product([cs.MX, cs.SX], [(1, 1), (3, 1), (1, 3), (3, 3)]))
    def test_cs2array_array2cs__convert_properly(
        self, sym_type: Union[type[cs.SX], type[cs.MX]], shape: tuple[int, int]
    ):
        x = sym_type.sym("x", *shape)
        a = cs2array(x)
        y = array2cs(a)
        for i in np.ndindex(shape):
            if sym_type is cs.SX:
                self.assertTrue(cs.is_equal(x[i], a[i]))
                self.assertTrue(cs.is_equal(x[i], y[i]))
            else:
                x_ = cs.DM(np.random.rand(*x.shape))
                o = cs.evalf(cs.substitute(x[i] - a[i], x, x_))
                np.testing.assert_allclose(o, 0, atol=1e-9)
                o = cs.evalf(cs.substitute(y[i] - a[i], x, x_))
                np.testing.assert_allclose(o, 0, atol=1e-9)

    @parameterized.expand([(cs.SX,), (cs.MX,)])
    def test_find_index_in_vector__works_properly(
        self, sym_type: Union[type[cs.SX], type[cs.MX]]
    ):
        X = sym_type.sym("X", 5)
        x = cs.horzcat(X[3], X[2])
        np.testing.assert_array_equal([3, 2], find_index_in_vector(X, x))


class TestDerivatives(unittest.TestCase):
    @parameterized.expand([((2, 2),), ((3, 1),), ((1, 3),)])
    def test_hojacobian__computes_right_derivatives(self, shape: tuple[int, int]):
        x = cs.SX.sym("x", *shape)
        y = (
            (x.reshape((-1, 1)) @ x.reshape((1, -1)) + (x.T if x.is_row() else x))
            if x.is_vector()
            else (x * x.T - x)
        )
        J = hojacobian(y, x)
        self.assertEqual(J.ndim, 4)
        for index in np.ndindex(J.shape):
            x_ = np.random.randn(*x.shape)
            idx1, idx2 = index[:2], index[2:]
            o = cs.evalf(cs.substitute(J[index] - cs.jacobian(y[idx1], x[idx2]), x, x_))
            np.testing.assert_allclose(o, 0, atol=1e-9)

    @parameterized.expand([((2, 2),), ((3, 1),), ((1, 3),)])
    def test_hohessian__computes_right_derivatives(self, shape: tuple[int, int]):
        x = cs.SX.sym("x", *shape)
        y = (
            (x.reshape((-1, 1)) @ x.reshape((1, -1)) + (x.T if x.is_row() else x))
            if x.is_vector()
            else (x * x.T - x)
        )
        H, _ = hohessian(y, x)
        self.assertEqual(H.ndim, 6)
        for i in np.ndindex(y.shape):
            x_ = np.random.randn(*x.shape)
            H_ = cs.hessian(y[i], x)[0]
            diff = array2cs(H[i].reshape(H_.shape, order="F")) - H_
            o = cs.evalf(cs.substitute(diff, x, x_))
            np.testing.assert_allclose(o, 0, atol=1e-9)


class TestScaling(unittest.TestCase):
    def test_str_and_repr(self):
        N = Scaler({"x": [-1, 1]})
        for S in [N.__str__(), N.__repr__()]:
            self.assertIn(Scaler.__name__, S)

    def test_register__raises__when_registering_duplicate_ranges(self):
        N = Scaler({"x": [-1, 1]})
        with self.assertRaises(KeyError):
            N.register("x", 0, 2)
        with self.assertRaises(KeyError):
            N.register("x", 0, 2)
        N.register("y", 0, 2)

    def test_can_scale__only_valid_ranges(self):
        N = Scaler({"x": [-1, 1]})
        self.assertTrue(N.can_scale("x"))
        self.assertFalse(N.can_scale("u"))

    def test_scaling__scale_unscale__computes_right_values(self):
        N = Scaler({"x1": (-5, 2), "x2": ([-1, 7], [2, 10])})
        x1 = 4
        y1 = N.scale("x1", x1)
        z1 = N.unscale("x1", y1)
        np.testing.assert_equal(y1, (4 + 5) / 2)
        np.testing.assert_equal(x1, z1)

        x2 = np.asarray([-5, 2])
        y2 = N.scale("x2", x2)
        z2 = N.unscale("x2", y2)
        np.testing.assert_equal(y2, [(-5 + 1) / 2, (2 - 7) / 10])
        np.testing.assert_equal(x2, z2)

    def test_minmaxscaling_scale_unscale__computes_right_values(self):
        N = MinMaxScaler(
            {"x1": (-5, 2), "x2": (np.asarray([-1, 7]), np.asarray([2, 10]))}
        )
        x1 = 4
        y1 = N.scale("x1", x1)
        z1 = N.unscale("x1", y1)
        np.testing.assert_equal(y1, (4 + 5) / (5 + 2))
        np.testing.assert_equal(x1, z1)

        x2 = np.asarray([-5, 2])
        y2 = N.scale("x2", x2)
        z2 = N.unscale("x2", y2)
        np.testing.assert_equal(y2, [(-5 + 1) / (2 + 1), (2 - 7) / (10 - 7)])
        np.testing.assert_equal(x2, z2)


if __name__ == "__main__":
    unittest.main()



import math as _math
import os
import pickle
import tempfile
import unittest
from itertools import product
from typing import Any, Optional

import casadi as cs
import numpy as np
from parameterized import parameterized
from scipy.stats import norm

from csnlp.core.solutions import subsevalf
from csnlp.util import io, math

TMPFILENAME: str = ""


class EmptyClass(io.SupportsDeepcopyAndPickle):
    ...


class DictClass(EmptyClass):
    def __init__(self) -> None:
        super().__init__()
        self.z, self.w = 3, 4
        self.sym2 = cs.SX.sym("y", 3, 1)


class TestIo(unittest.TestCase):
    @parameterized.expand(
        [
            (5.0, False),
            (unittest.TestCase(), False),
            (cs.DM(5), True),
            (cs.SX.sym("x"), True),
            (cs.MX.sym("x"), True),
        ]
    )
    def test_is_casadi_object__guesses_correctly(self, obj: Any, result: bool):
        self.assertEqual(io.is_casadi_object(obj), result)

    def test_is_pickleable__fails_with_casadi_obj(self):
        self.assertTrue(io.is_pickleable(5))
        self.assertTrue(io.is_pickleable({5}))
        self.assertTrue(io.is_pickleable("hello"))
        self.assertFalse(io.is_pickleable(cs.SX.sym("x")))

    @parameterized.expand(
        [
            ("pkl",),
            ("xz",),
            ("pbz2",),
            ("gz",),
            ("bt",),
            ("bl2",),
            ("mat",),
            ("npz",),
        ]
    )
    def test_save_and_load__preserve_data_correctly(self, ext: str):
        global TMPFILENAME
        TMPFILENAME = f"{next(tempfile._get_candidate_names())}.{ext}"
        data = {"x": 5, "y": "ciao", "w": {"ci": "ao"}}
        io.save(TMPFILENAME, **data)
        data2 = io.load(TMPFILENAME)
        self.assertDictEqual(data, data2)

    def test_save_and_load__one_key_dict_is_simplified(self):
        global TMPFILENAME
        TMPFILENAME = f"{next(tempfile._get_candidate_names())}.pkl"
        data = {"x": 5}
        io.save(TMPFILENAME, **data)
        data2 = io.load(TMPFILENAME)
        self.assertIsNot(data2, dict)
        self.assertEqual(data["x"], data2)

    def tearDown(self) -> None:
        try:
            os.remove(TMPFILENAME)
        finally:
            return super().tearDown()

    @parameterized.expand([(False,), (True,)])
    def test_is_pickleable_and_deepcopy_able(self, copy: bool):
        ec = EmptyClass()
        dc = DictClass()
        if copy:
            ec1 = ec.copy()
            dc1 = dc.copy()
        else:
            ec1 = pickle.loads(pickle.dumps(ec))
            dc1 = pickle.loads(pickle.dumps(dc))
        self.assertIsNot(ec, ec1)
        self.assertIsNot(dc, dc1)
        self.assertTupleEqual((dc.z, dc.w), (dc1.z, dc1.w))
        if copy:
            self.assertTupleEqual(dc.sym2.shape, dc1.sym2.shape)
        else:
            self.assertFalse(hasattr(dc1, "sym2"))


class TestMath(unittest.TestCase):
    def test_log(self):
        base = np.random.rand() * 10
        x = np.random.rand() * 10
        self.assertEqual(math.log(x), _math.log(x))
        self.assertEqual(math.log(x, base), _math.log(x, base))

    @parameterized.expand(product([(3, 1), (1, 3), (3, 4)], [-2, -1, 0, 1, None]))
    def test_prod(self, shape: tuple[int, int], axis: Optional[int]):
        x = np.random.randn(*shape) * 3
        p = np.prod(x, axis=axis, keepdims=True)
        x_sx = cs.SX.sym("X", *shape)
        p_sx = subsevalf(math.prod(x_sx, axis=axis), x_sx, x)
        x_mx = cs.MX.sym("X", *shape)
        p_mx = subsevalf(math.prod(x_mx, axis=axis), x_mx, x)
        x_dm = cs.DM(x)
        p_dm = math.prod(x_dm, axis=axis)
        np.testing.assert_allclose(p, p_sx)
        np.testing.assert_allclose(p, p_mx)
        np.testing.assert_allclose(p, p_dm)

    def test_norm_cdf(self):
        shape = (3, 4)
        x_sx = cs.SX.sym("X", *shape)
        loc_sx = cs.SX.sym("loc", *shape)
        scale_sx = cs.SX.sym("scale", *shape)
        x_mx = cs.MX.sym("X", *shape)
        loc_mx = cs.MX.sym("loc", *shape)
        scale_mx = cs.MX.sym("scale", *shape)
        x = np.random.randn(*shape)
        loc = np.random.randn(*shape)
        scale = np.random.rand(*shape)
        cdf_sx = subsevalf(
            math.norm_cdf(x_sx, loc=loc_sx, scale=scale_sx),
            [x_sx, loc_sx, scale_sx],
            [x, loc, scale],
        )
        cdf_mx = subsevalf(
            math.norm_cdf(x_mx, loc=loc_mx, scale=scale_mx),
            [x_mx, loc_mx, scale_mx],
            [x, loc, scale],
        )
        cdf = norm.cdf(x, loc=loc, scale=scale)

        np.testing.assert_allclose(cdf, cdf_sx, atol=1e-7, rtol=1e-5)
        np.testing.assert_allclose(cdf, cdf_mx, atol=1e-7, rtol=1e-5)

    def test_norm_ppf(self):
        shape = (3, 4)
        x_sx = cs.SX.sym("X", *shape)
        loc_sx = cs.SX.sym("loc", *shape)
        scale_sx = cs.SX.sym("scale", *shape)
        x_mx = cs.MX.sym("X", *shape)
        loc_mx = cs.MX.sym("loc", *shape)
        scale_mx = cs.MX.sym("scale", *shape)
        x = np.random.rand(*shape)
        loc = np.random.randn(*shape)
        scale = np.random.rand(*shape)
        cdf_sx = subsevalf(
            math.norm_ppf(x_sx, loc=loc_sx, scale=scale_sx),
            [x_sx, loc_sx, scale_sx],
            [x, loc, scale],
        )
        cdf_mx = subsevalf(
            math.norm_ppf(x_mx, loc=loc_mx, scale=scale_mx),
            [x_mx, loc_mx, scale_mx],
            [x, loc, scale],
        )
        cdf = norm.ppf(x, loc=loc, scale=scale)

        np.testing.assert_allclose(cdf, cdf_sx, atol=1e-5, rtol=1e-5)
        np.testing.assert_allclose(cdf, cdf_mx, atol=1e-5, rtol=1e-5)

    @parameterized.expand(
        [
            ((3, 4), cs.DM([3, 3, 3, 3])),
            (([1, 2, 3, 4], 3), cs.DM(sum(([i] * 3 for i in range(1, 5)), []))),
            ((cs.DM([[1, 2], [3, 4]]), (1, 2)), cs.DM([[1, 1, 2, 2], [3, 3, 4, 4]])),
        ]
    )
    def test_repeat(self, inputs, expected):
        actual = math.repeat(*inputs)
        np.testing.assert_array_equal(actual, expected)


if __name__ == "__main__":
    unittest.main()



import pickle
import unittest
import warnings

import casadi as cs
import numpy as np
from parameterized import parameterized, parameterized_class

from csnlp import Nlp, scaling
from csnlp.core.solutions import subsevalf
from csnlp.util.math import log
from csnlp.wrappers import (
    Mpc,
    NlpScaling,
    NlpSensitivity,
    NonRetroactiveWrapper,
    Wrapper,
)

OPTS = {
    "expand": True,
    "print_time": False,
    "ipopt": {
        "max_iter": 500,
        "tol": 1e-7,
        "barrier_tol_factor": 5,
        "sb": "yes",
        # for debugging
        "print_level": 0,
        "print_user_options": "no",
        "print_options_documentation": "no",
    },
}


class TestWrapper(unittest.TestCase):
    def test_attr__raises__when_accessing_private_attrs(self):
        nlp = Nlp()
        nlp._x
        wrapped = Wrapper[cs.SX](nlp)
        with self.assertRaisesRegex(
            AttributeError, "Accessing private attribute '_x' is prohibited."
        ):
            wrapped._x

    def test_unwrapped__unwraps_nlp_correctly(self):
        nlp = Nlp()
        self.assertIs(nlp, nlp.unwrapped)
        wrapped = Wrapper[cs.SX](nlp)
        self.assertIs(nlp, wrapped.unwrapped)

    def test_str_and_repr(self):
        nlp = Nlp()
        wrapped = Wrapper[cs.SX](nlp)
        S = wrapped.__str__()
        self.assertIn(Wrapper.__name__, S)
        self.assertIn(nlp.__str__(), S)
        S = wrapped.__repr__()
        self.assertIn(Wrapper.__name__, S)
        self.assertIn(nlp.__repr__(), S)

    def test_is_wrapped(self):
        nlp = Nlp()
        self.assertFalse(nlp.is_wrapped())

        wrapped = Mpc[cs.SX](nlp=nlp, prediction_horizon=10)
        self.assertTrue(wrapped.is_wrapped(Mpc))
        self.assertFalse(wrapped.is_wrapped(NlpSensitivity))

        wrapped = NlpSensitivity[cs.SX](nlp=nlp)
        self.assertFalse(wrapped.is_wrapped(Mpc))
        self.assertTrue(wrapped.is_wrapped(NlpSensitivity))

        wrapped = NlpSensitivity[cs.SX](nlp=Mpc(nlp=nlp, prediction_horizon=10))
        self.assertTrue(wrapped.is_wrapped(Mpc))
        self.assertTrue(wrapped.is_wrapped(NlpSensitivity))

        wrapped_pickled = pickle.loads(pickle.dumps(wrapped))
        wrapped_copied = wrapped.copy()

        self.assertEqual(wrapped.name, wrapped_pickled.name)
        self.assertEqual(repr(wrapped), repr(wrapped_pickled))
        self.assertEqual(wrapped.name, wrapped_copied.name)
        self.assertEqual(str(wrapped), str(wrapped_copied))
        self.assertEqual(repr(wrapped), repr(wrapped_copied))


class TestNonRetroactiveWrapper(unittest.TestCase):
    def test_init__raises__with_variable_already_defined(self):
        nlp = Nlp()
        nlp.variable("x")
        with self.assertRaisesRegex(ValueError, "Nlp already defined."):
            NonRetroactiveWrapper(nlp)

    def test_init__raises__with_parameter_already_defined(self):
        nlp = Nlp()
        nlp.parameter("p")
        with self.assertRaisesRegex(ValueError, "Nlp already defined."):
            NonRetroactiveWrapper(nlp)


@parameterized_class("sym_type", [("SX",), ("MX",)])
class TestNlpSensitivity(unittest.TestCase):
    def test_lagrangian__is_correct__example_1a_b(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        _, lam = nlp.constraint("c1", x**2 + y**2, "==", 1)
        for f in [-x - y, (x + y) ** 2]:
            nlp.minimize(f)
            L = f + lam * (x**2 + y**2 - 1)
            x_ = np.random.randn(*x.shape)
            y_ = np.random.randn(*y.shape)
            lam_ = np.random.randn(*lam.shape)
            np.testing.assert_allclose(
                subsevalf(nlp.lagrangian, [x, y, lam], [x_, y_, lam_]),
                subsevalf(L, [x, y, lam], [x_, y_, lam_]),
            )

    def test_lagrangian__is_correct__example_3(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_3
        n = 50
        lbx = 1 / n * 1e-6
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        p, lam_lbx, _ = nlp.variable("p", (n, 1), lb=lbx)
        c1, lam_g = nlp.constraint("c1", cs.sum1(p), "==", 1)
        f = cs.sum1(p * log(p, 2))
        nlp.minimize(f)
        L = f + lam_g * c1 + lam_lbx.T @ (lbx - p)
        p_ = np.random.rand(*p.shape)
        lam_lbx_ = np.random.randn(*lam_lbx.shape)
        lam_g_ = np.random.randn(*lam_g.shape)
        np.testing.assert_allclose(
            subsevalf(nlp.lagrangian, [p, lam_lbx, lam_g], [p_, lam_lbx_, lam_g_]),
            subsevalf(L, [p, lam_lbx, lam_g], [p_, lam_lbx_, lam_g_]),
        )

    def test_lagrangian__is_correct__example_5(self):
        # https://personal.math.ubc.ca/~israel/m340/kkt2.pdf
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x, lam_lbx, lam_ubx = nlp.variable("x", (2, 1), lb=0, ub=1)
        c1, lam_h = nlp.constraint("c1", x[0] + x[1] ** 2, "<=", 2)
        f = -x[0] * x[1]
        nlp.minimize(f)
        L = f + lam_h * c1 + lam_lbx.T @ (0 - x) + lam_ubx.T @ (x - 1)
        x_ = np.random.rand(*x.shape)
        lam_lbx_ = np.random.randn(*lam_lbx.shape)
        lam_ubx_ = np.random.randn(*lam_ubx.shape)
        lam_h_ = np.random.randn(*lam_h.shape)
        np.testing.assert_allclose(
            subsevalf(
                nlp.lagrangian,
                [x, lam_lbx, lam_ubx, lam_h],
                [x_, lam_lbx_, lam_ubx_, lam_h_],
            ),
            subsevalf(
                L, [x, lam_lbx, lam_ubx, lam_h], [x_, lam_lbx_, lam_ubx_, lam_h_]
            ),
        )

    @parameterized.expand([(False,), (True,)])
    def test_kkt__returns_tau_correctly(self, flag: bool):
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type), include_barrier_term=flag)
        x = nlp.variable("x")[0]
        nlp.constraint("c", x, "<=", 1)
        nlp.minimize(x)
        _, tau = nlp.kkt
        if flag:
            self.assertIsInstance(tau, nlp.unwrapped.sym_type)
        else:
            self.assertIsNone(tau)

    def test_kkt__computes_kkt_conditions_correctly__example_1a(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 1)
        nlp.minimize(-x - y)
        nlp.init_solver(OPTS)
        sol = nlp.solve()
        kkt, _ = nlp.kkt
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-9)

    def test_kkt__computes_kkt_conditions_correctly__example_1b(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_1
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 1)
        nlp.minimize((x + y) ** 2)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": 0.5, "y": np.sqrt(1 - 0.5**2)})
        kkt, _ = nlp.kkt
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-6)

    def test_kkt__computes_kkt_conditions_correctly__example_2(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_2
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x")[0]
        y = nlp.variable("y")[0]
        nlp.constraint("c1", x**2 + y**2, "==", 3)
        nlp.minimize(x**2 * y)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": np.sqrt(3 - 0.8**2), "y": -0.8})
        kkt, _ = nlp.kkt
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-6)

    def test_kkt__computes_kkt_conditions_correctly__example_3(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_3
        n = 50
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        p = nlp.variable("p", (n, 1), lb=1 / n * 1e-6)[0]
        nlp.constraint("c1", cs.sum1(p), "==", 1)
        nlp.minimize(cs.sum1(p * log(p, 2)))
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"p": np.random.rand(n)})
        kkt, tau = nlp.kkt
        kkt = subsevalf(kkt, tau, sol.barrier_parameter, eval=False)
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-6)

    def test_kkt__computes_kkt_conditions_correctly__example_4(self):
        # https://en.wikipedia.org/wiki/Lagrange_multiplier#Example_4
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x")[0]
        nlp.constraint("c1", x**2, "==", 1)
        nlp.minimize(x**2)
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": 1 + 0.1 * np.random.rand()})
        kkt, _ = nlp.kkt
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-6)

    def test_kkt__computes_kkt_conditions_correctly__example_5(self):
        # https://personal.math.ubc.ca/~israel/m340/kkt2.pdf
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x", (2, 1), lb=0)[0]
        nlp.constraint("c1", x[0] + x[1] ** 2, "<=", 2)
        nlp.minimize(-x[0] * x[1])
        nlp.init_solver(OPTS)
        sol = nlp.solve(vals0={"x": [1, 1]})
        kkt, tau = nlp.kkt
        kkt = subsevalf(kkt, tau, sol.barrier_parameter, eval=False)
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-7)

    def test_kkt__computes_sensitivity_correctly__example_7(self):
        #  Example 4.2 from [1]
        #
        # References
        # ----------
        # [1] Buskens, C. and Maurer, H. (2001). Sensitivity analysis and
        #     real-time optimization of parametric nonlinear programming
        #     problems. In M. Grotschel, S.O. Krumke, and J. Rambau (eds.),
        #     Online Optimization of Large Scale Systems, 3–16. Springer,
        #     Berlin, Heidelberg.
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        z = nlp.variable("z", (2, 1))[0]
        p = nlp.parameter("p")
        nlp.minimize(-((0.5 + p) * cs.sqrt(z[0]) + (0.5 - p) * z[1]))
        _, lam = nlp.constraint("c1", cs.sum1(z), "<=", 1)
        nlp.constraint("c2", z[0], ">=", 0.1)
        nlp.init_solver(OPTS)
        sol = nlp.solve(pars={"p": 0}, vals0={"z": [0.1, 0.9]})

        np.testing.assert_allclose(sol.vals["z"], [[0.25], [0.75]], rtol=1e-5)
        np.testing.assert_allclose(sol.value(lam), 0.5)

        kkt, tau = nlp.kkt
        kkt = subsevalf(kkt, tau, sol.barrier_parameter, eval=False)
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-7)

        S1 = sol.value(nlp.parametric_sensitivity()[0]).full().flatten()
        S2 = nlp.parametric_sensitivity(solution=sol)[0].flatten()
        for S in (S1, S2):
            np.testing.assert_allclose(S, [2, -2, -1, 0], atol=1e-5)

    def test_kkt__computes_sensitivity_correctly__example_8(self):
        #  Example 4.5 from [1]
        #
        # References
        # ----------
        # [1] Buskens, C. and Maurer, H. (2001). Sensitivity analysis and
        #     real-time optimization of parametric nonlinear programming
        #     problems. In M. Grotschel, S.O. Krumke, and J. Rambau (eds.),
        #     Online Optimization of Large Scale Systems, 3–16. Springer,
        #     Berlin, Heidelberg.
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        z = nlp.variable("z", (2, 1))[0]
        p = nlp.parameter("p")
        nlp.minimize(cs.sumsqr(z + [1, -2]))
        _, lam1 = nlp.constraint("c1", -z[0] + p, "<=", 0)
        _, lam2 = nlp.constraint("c2", 2 * z[0] + z[1], "<=", 6)
        nlp.init_solver(OPTS)
        sol = nlp.solve(pars={"p": 1})

        np.testing.assert_allclose(sol.f, 4)
        np.testing.assert_allclose(sol.vals["z"], [[1], [2]], rtol=1e-7)
        np.testing.assert_allclose(sol.value(lam1), 4, atol=1e-7)
        np.testing.assert_allclose(sol.value(lam2), 0, atol=1e-7)

        kkt, tau = nlp.kkt
        kkt = subsevalf(kkt, tau, sol.barrier_parameter, eval=False)
        np.testing.assert_allclose(sol.value(kkt), 0, atol=1e-7)

        S1 = nlp.parametric_sensitivity()[0]
        S2 = nlp.parametric_sensitivity(solution=sol)[0]
        np.testing.assert_allclose(sol.value(S1).full().flat, [1, 0, 2, 0], atol=1e-5)
        np.testing.assert_allclose(S2.flat, [1, 0, 2, 0], atol=1e-5)

        Fp1, Fpp1 = (
            sol.value(o)
            for o in nlp.parametric_sensitivity(expr=nlp.f, second_order=True)
        )
        Fp2, Fpp2 = nlp.parametric_sensitivity(
            expr=nlp.f, solution=sol, second_order=True
        )
        np.testing.assert_allclose(Fp1, 4, atol=1e-7)
        np.testing.assert_allclose(Fp2, 4, atol=1e-7)
        np.testing.assert_allclose(Fpp1, 2, atol=1e-7)
        np.testing.assert_allclose(Fpp2, 2, atol=1e-7)

    def test_licq__computes_qualification_correctly__example_1(self):
        # https://de.wikipedia.org/wiki/Linear_independence_constraint_qualification#LICQ
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x", (2, 1))[0]
        nlp.constraint("c1", cs.sum1(x), "<=", 1)
        nlp.constraint("c2", cs.sumsqr(x), "<=", 1)
        nlp.minimize(x[0] ** 2 + (x[1] - 1) ** 2)  # any objective

        x_ = cs.DM([[0], [1]])
        d = subsevalf(nlp.licq, x, x_).full()
        np.testing.assert_allclose(d, [[1, 1], [0, 2]])
        self.assertEqual(np.linalg.matrix_rank(d), d.shape[0])

    def test_licq__computes_qualification_correctly__example_2(self):
        # https://de.wikipedia.org/wiki/Linear_independence_constraint_qualification#MFCQ_ohne_LICQ
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x", (2, 1))[0]
        nlp.constraint("c1", x[1], ">=", 0)
        nlp.constraint("c2", x[0] ** 4 - x[1], "<=", 1)
        nlp.minimize(x[0] ** 2 + x[1] ** 2)  # any objective

        x_ = cs.DM([[0], [0]])
        d = subsevalf(nlp.licq, x, x_).full()
        np.testing.assert_allclose(d, [[0, -1], [0, -1]])
        self.assertNotEqual(np.linalg.matrix_rank(d), d.shape[0])

    def test_parametric_sensitivity__is_correct(self):
        # https://web.casadi.org/blog/nlp_sens/
        p_values_and_solutions = [
            (1.25, (-0.442401495488, 0.400036517837, 3.516850030791)),
            (1.4, (-0.350853791920, 0.766082818877, 1.401822559490)),
            (2, (-0.000000039328, 0, 0)),
        ]

        def z(x):
            return x[1, :] - x[0, :]

        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x", (2, 1), lb=[[0], [-np.inf]])[0]
        p = nlp.parameter("p", (2, 1))
        nlp.minimize((1 - x[0]) ** 2 + p[0] * (x[1] - x[0] ** 2) ** 2)
        g = (x[0] + 0.5) ** 2 + x[1] ** 2
        nlp.constraint("c1", (p[1] / 2) ** 2, "<=", g)
        nlp.constraint("c2", g, "<=", p[1] ** 2)
        nlp.init_solver(OPTS)

        Z1_ = z(x)
        J1_, H1_ = nlp.parametric_sensitivity(expr=Z1_, second_order=True)
        for p, (Z, J, H) in p_values_and_solutions:
            sol = nlp.solve(pars={"p": [0.2, p]})
            Z1 = sol.value(Z1_)
            J1 = sol.value(J1_)
            H1 = sol.value(H1_)
            J2, H2 = nlp.parametric_sensitivity(
                expr=Z1_, solution=sol, second_order=True
            )
            J2, H2 = np.squeeze(J2), np.squeeze(H2)
            np.testing.assert_allclose(J1.full().flat, J2, atol=1e-7)
            np.testing.assert_allclose(H1.full(), H2, atol=1e-7)
            np.testing.assert_allclose(Z, Z1, atol=1e-4)
            np.testing.assert_allclose(J, J1[1], atol=1e-4)
            np.testing.assert_allclose(J, J2[1], atol=1e-4)
            np.testing.assert_allclose(H, H1[1, 1], atol=1e-4)
            np.testing.assert_allclose(H, H2[1, 1], atol=1e-4)

    def test_parametric_sensitivity__with_specific_pars__is_correct(self):
        # https://web.casadi.org/blog/nlp_sens/
        p_values_and_solutions = [
            (1.25, (-0.442401495488, 0.400036517837, 3.516850030791)),
            (1.4, (-0.350853791920, 0.766082818877, 1.401822559490)),
            (2, (-0.000000039328, 0, 0)),
        ]

        def z(x):
            return x[1, :] - x[0, :]

        nlp = Nlp(sym_type=self.sym_type)
        x = nlp.variable("x", (2, 1), lb=[[0], [-np.inf]])[0]
        p = nlp.parameter("p", (2, 1))
        p2 = nlp.parameter("p2", (1, 1))
        nlp.minimize((1 - x[0]) ** 2 + p[0] * (x[1] - x[0] ** 2) ** 2)
        g = (x[0] + 0.5) ** 2 + x[1] ** 2 + p2
        nlp.constraint("c1", (p[1] / 2) ** 2, "<=", g)
        nlp.constraint("c2", g, "<=", p[1] ** 2)
        nlp.init_solver(OPTS)

        sensitivity_all = NlpSensitivity(nlp)
        sensitivity_partial = NlpSensitivity(nlp, target_parameters=p[1])

        kwargs = dict(expr=z(x), second_order=True)
        J1_, H1_ = sensitivity_all.parametric_sensitivity(**kwargs)
        J2_, H2_ = sensitivity_partial.parametric_sensitivity(**kwargs)
        for p, (_, J_exp, H_exp) in p_values_and_solutions:
            sol = nlp.solve(pars={"p": [0.2, p], "p2": 0})
            J1_act = sol.value(J1_)
            H1_act = sol.value(H1_)
            J2_act = sol.value(J2_)
            H2_act = sol.value(H2_)
            np.testing.assert_allclose(J_exp, J1_act[1], atol=1e-4)
            np.testing.assert_allclose(H_exp, H1_act[1, 1], atol=1e-4)
            np.testing.assert_allclose(J_exp, J2_act, atol=1e-4)
            np.testing.assert_allclose(H_exp, H2_act, atol=1e-4)

    def test_can_be_pickled(self):
        nlp = NlpSensitivity(Nlp(sym_type=self.sym_type))
        x = nlp.variable("x", (2, 1), lb=[[0], [-np.inf]])[0]
        p = nlp.parameter("p", (2, 1))
        nlp.minimize((1 - x[0]) ** 2 + p[0] * (x[1] - x[0] ** 2) ** 2)
        g = (x[0] + 0.5) ** 2 + x[1] ** 2
        nlp.constraint("c1", (p[1] / 2) ** 2, "<=", g)
        nlp.constraint("c2", g, "<=", p[1] ** 2)
        nlp.init_solver(OPTS)

        nlp2 = pickle.loads(pickle.dumps(nlp))

        self.assertIn(repr(nlp), repr(nlp2))


class TestNlpScaling(unittest.TestCase):
    @parameterized.expand([("variable",), ("parameter",)])
    def test_parameter_and_variable__warns__when_cannot_scale(self, method: str):
        scaler = scaling.Scaler()
        nlp = NlpScaling[cs.SX](Nlp(sym_type="SX"), scaler=scaler, warns=True)
        with warnings.catch_warnings():
            warnings.filterwarnings("error")
            with self.assertRaisesRegex(
                RuntimeWarning, f"Scaling for {method} p not found."
            ):
                getattr(nlp, method)("p")

    def test_parameter__scales_correctly(self):
        scaler = scaling.Scaler({"p": (0, 2)})
        nlp = NlpScaling[cs.SX](Nlp(sym_type="SX"), scaler=scaler, warns=True)
        p = nlp.parameter("p")
        self.assertIn("p", nlp.scaled_parameters)
        self.assertIn("p", nlp.unscaled_parameters)
        p_s = nlp.scale(p)
        p_u = nlp.unscale(p)
        np.testing.assert_allclose(cs.evalf(nlp.scaled_parameters["p"] - p_s), 0)
        np.testing.assert_allclose(cs.evalf(nlp.unscaled_parameters["p"] - p_u), 0)

    def test_variable__scales_correctly(self):
        scaler = scaling.Scaler({"x": (0, 2)})
        nlp = NlpScaling[cs.SX](Nlp(sym_type="SX"), scaler=scaler, warns=True)
        lb, ub = -5, 4
        x, _, _ = nlp.variable("x", lb=lb, ub=ub)
        self.assertIn("x", nlp.scaled_variables)
        self.assertIn("x", nlp.unscaled_variables)
        x_s = nlp.scale(x)
        x_u = nlp.unscale(x)
        np.testing.assert_allclose(cs.evalf(nlp.scaled_variables["x"] - x_s), 0)
        np.testing.assert_allclose(cs.evalf(nlp.unscaled_variables["x"] - x_u), 0)
        np.testing.assert_allclose(nlp.lbx, scaler.scale("x", lb))
        np.testing.assert_allclose(nlp.ubx, scaler.scale("x", ub))


if __name__ == "__main__":
    unittest.main()



from math import sqrt
from typing import Optional, Union

import casadi as cs

SQRT2 = sqrt(2)


def log(
    x: Union[cs.SX, cs.MX, cs.DM], base: Union[None, cs.SX, cs.MX, cs.DM] = None
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Logarithm. With one argument, return the natural logarithm of `x` (to base `e`).
    With two arguments, return the logarithm of x to the given base, calculated as
    `log(x) / log(base)`.

    Parameters
    ----------
    x : Union[cs.SX, cs.MX, cs.DM]
        Value to compute the logarithm of.
    base : Union[None, cs.SX, cs.MX, cs.DM], optional
        Base of the logarithm, by default `None`

    Returns
    -------
    Union[cs.SX, cs.MX, cs.DM]
        The logarithm. of `x` with base `base`.
    """
    if base is None:
        return cs.log(x)
    if base == 10 or base == 10.0:
        return cs.log10(x)
    return cs.log(x) / cs.log(base)


def prod(
    x: Union[cs.SX, cs.MX, cs.DM], axis: Optional[int] = None
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Computes the product of all the elements in `x` (CasADi version of `numpy.prod`).

    Parameters
    ----------
    x : Union[cs.SX, cs.MX, cs.DM]
        The variable whose entries must be multiplied together.
    axis : {0, 1, None, -1, -2}
        Axis or along which a product is performed. The default, `axis=None`, will
        calculate the product of all the elements in the matrix. If axis is negative it
        counts from the last to the first axis.

    Returns
    -------
    Union[cs.SX, cs.MX, cs.DM]
        Product of the elements.
    """
    # sourcery skip: merge-comparisons
    if axis is None:
        x = cs.vec(x)
        axis = 0

    # prod of vector elements
    if x.is_vector():
        if x.shape[axis] == 1:
            return x
        if not isinstance(x, cs.MX):
            return cs.det(cs.diag(x))  # does not work with MX
        p = x[0]
        for i in range(1, x.shape[axis]):
            p *= x[i]
        return p

    # prod of matrix elements
    sum_ = cs.sum1 if (axis == 0 or axis == -2) else cs.sum2
    n_negatives = sum_(x < 0)
    p = cs.exp(sum_(cs.log(cs.fabs(x))))
    return cs.if_else(cs.remainder(n_negatives, 2) == 0.0, 1, -1) * p


def norm_cdf(
    x: Union[cs.SX, cs.MX, cs.DM],
    loc: Union[cs.SX, cs.MX, cs.DM] = cs.DM(0),
    scale: Union[cs.SX, cs.MX, cs.DM] = cs.DM(1),
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Computes the cdf of a normal distribution. See `scipy.stats.norm.cdf`.

    Parameters
    ----------
    x : Union[cs.SX, cs.MX, cs.DM]
        The value at which the cdf is computed.
    loc : Union[cs.SX, cs.MX, cs.DM], optional
        Mean of the normal distribution. By default, `loc=0`.
    scale : Union[cs.SX, cs.MX, cs.DM], optional
        Standard deviation of the normal distribution. By default, `scale=0`.

    Returns
    -------
    Union[cs.SX, cs.MX, cs.DM]
        The cdf of the normal distribution.
    """
    return 0.5 * (1 + cs.erf((x - loc) / SQRT2 / scale))


def norm_ppf(
    p: Union[cs.SX, cs.MX, cs.DM],
    loc: Union[cs.SX, cs.MX, cs.DM] = cs.DM(0),
    scale: Union[cs.SX, cs.MX, cs.DM] = cs.DM(1),
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Computes the quantile (invese of cdf) of a normal distribution. See
    `scipy.stats.norm.ppf`.

    Parameters
    ----------
    x : Union[cs.SX, cs.MX, cs.DM]
        The value at which the quantile is computed.
    loc : Union[cs.SX, cs.MX, cs.DM], optional
        Mean of the normal distribution. By default, `loc=0`.
    scale : Union[cs.SX, cs.MX, cs.DM], optional
        Standard deviation of the normal distribution. By default, `scale=0`.

    Returns
    -------
    Union[cs.SX, cs.MX, cs.DM]
        The quantile of the normal distribution.
    """
    return SQRT2 * scale * cs.erfinv(2 * p - 1) + loc


def repeat(
    a: Union[cs.SX, cs.MX, cs.DM], repeats: Union[int, tuple[int, int]]
) -> Union[cs.SX, cs.MX, cs.DM]:
    """Repeats elements in array.

    Parameters
    ----------
    a : casadi.SX or MX or DM
        The array/matrix whose elements are to be repeated.
    repeats : int or tuple[int, int]
        The number of repeats, in the first and/or second axis.

    Returns
    -------
    Union[cs.SX, cs.MX, cs.DM]
        Output array with repeated elements.
    """
    return cs.kron(a, cs.GenDM_ones(repeats))
